/*
 * MemoryAdhoc.cpp
 *
 * Time-stamp: <04/09/09 16:07:10 lars>
 *
 * history: 2004/04/01  1.0  LST created.
 */

#include "Debug.h"
#include "MemoryAdhoc.h"
#include "Eclat.h"

#include <iostream>
using namespace std;
// #include <stdio.h>


/**
 * Compute the difference of every other row (with higher index) minus a given row.
 * I.e., compute bmIn_row' \setminus bmIn_row for all row' > row.
 * @param bmIn original bit matrix.
 * @param row row to compute rest by.
 * @param minsup minimal support.
 */
SparseBitmatrix* setdifferenceBySparseBitmatrix_dynamicMemory(SparseBitmatrix* bmIn, int row, int maxdefect) {
  assert(row >= 0 && row < bmIn->numRows);

#if DEBUG_LEVEL >= 20
  cout << "intersect by row1 = " << row << " of " << bmIn->numRows << " ... "; cout.flush();
  cout << endl << "bmIn:" << endl;
  bmIn->dump();
  cout << "dump done" << endl;
#endif

  // 1. reset structure of bmOut.
  int numRowsIn = bmIn->numRows;
  int** rowsIn = bmIn->rows;
  int* rowLengthsIn = bmIn->rowLengths;
  int numRowsOut = numRowsIn - row - 1;
  SparseBitmatrix* bmOut = new SparseBitmatrix(numRowsOut);
  assert(bmOut != NULL);
  assert(numRowsOut <= bmOut->sizeRows);
  int** rowsOut = bmOut->rows;
  int* rowLengthsOut = bmOut->rowLengths;

  // 2. do the intersection:
  int* row1 = rowsIn[row];
  assert(row1 != NULL);
  int len1 = rowLengthsIn[row];
  int maxRowLength = 0;
  int tor = 0;
  int* rowOut = NULL;

  if (len1 == 0) { // copy bmIn to bmOut:
    for (int fromr = row+1; fromr < numRowsIn; fromr++) {
      tor = fromr-row-1;
      int* row2 = rowsIn[fromr];
      assert(row2 != NULL);
      int len2 = rowLengthsIn[fromr];
      if (len2 > maxRowLength)
	maxRowLength = len2;

      rowsOut[tor] = new int[len2]; bmOut->rowSizes[tor] = len2;
      bmOut->rowLengths[tor] = len2;
      rowOut = rowsOut[tor];
      assert(rowOut != NULL);
      assert(rowsOut[tor] != NULL);
      assert(bmOut->rowSizes[tor] >= len2 || bmOut->rowSizes[tor] >= maxdefect);
      for (int pos = 0; pos < len2; pos++)
	rowOut[pos] = row2[pos];
    }
    bmOut->numRows = tor;
    bmOut->sizeRows = tor;
    bmOut->maxRowLength = maxRowLength;
#if DEBUG_LEVEL >= 20
    cout << "- done (copy)" << endl;
#endif
    return bmOut;
  }

  for (int fromr = row+1; fromr < numRowsIn; fromr++) {
    assert(tor >= 0 && tor < numRowsOut);
#if DEBUG_LEVEL >= 25
    cout << "fromr = " << fromr << " of " << numRowsIn << endl;
#endif
    int* row2 = rowsIn[fromr];
    assert(row2 != NULL);
    int len2 = rowLengthsIn[fromr];
    // assert(bmOut->rowSizes[tor] >= len2);

    int maxLen = len2 < maxdefect ? len2 : maxdefect+1;
    assert(maxLen >= 0);
    rowsOut[tor] = new int[maxLen]; 
    assert(bmOut != NULL);
    bmOut->rowSizes[tor] = maxLen; 
    rowOut = rowsOut[tor];
    assert(rowOut != NULL);
    assert(rowsOut[tor] != NULL);
    assert(bmOut->rowSizes[tor] >= len2 || bmOut->rowSizes[tor] >= maxdefect);

    int maxdefectNew = maxdefect;
    int posOut = 0;

    if (len2 > 0) {
#if DEBUG_LEVEL >= 25
      cout << "  len2 = " << len2 << " > 0 " << endl;
#endif
      int pos1 = 0, pos2 = 0;
      // if (len1 == 0) { cout << "HHHHHHHHHHHHHHAAAAAAAAAA" << endl;}
      int item1 = row1[0];
      
      assert(fromr < bmIn->sizeRows);
      assert(row2 == bmIn->rows[fromr]);
      //assert(pos2 < bmIn->rowSizes[fromr]);
      assert(0 < bmIn->rowSizes[fromr]);
      
      assert(row2 != NULL);
      int item2 = row2[0];
      
#if DEBUG_LEVEL >= 25
      cout << "  start loop" << endl;
#endif
      // a. compute set difference:
      while (true) {
	if (item1 < item2) {
	  pos1++; 
	  if (pos1 >= len1) break;
	  item1 = row1[pos1];
	} else if (item1 > item2) {
	  assert(posOut >= 0);
	  assert(posOut < len2); 
	  assert(tor >=0 && tor < bmOut->sizeRows);
	  assert(posOut < bmOut->rowSizes[tor]);
	  rowOut[posOut++] = item2;
	  maxdefectNew--;
	  if (maxdefectNew < 0) break;
	  
	  pos2++; 
	  if (pos2 >= len2) break;
	  item2 = row2[pos2];
	} else { // if (item1 == item2) {
	  pos2++; if (pos2 >= len2) break; // first check for pos2 as we add all remaining row2 items at the end !
	  pos1++; if (pos1 >= len1) break;
	  item1 = row1[pos1]; item2 = row2[pos2];
	}
      }

#if DEBUG_LEVEL >= 25
      cout << "  end loop" << endl;
#endif

      // copy remaining items of row2:
      if (pos2 < len2 && maxdefectNew >= 0) {
	maxdefectNew -= (len2-pos2);
	if (maxdefectNew >= 0) {
	  while (pos2 < len2) {
	    assert(posOut < bmOut->rowSizes[tor]);
	    rowOut[posOut++] = row2[pos2++];
	  }
	}
      }

#if DEBUG_LEVEL >= 25
      cout << "  copied rest" << endl;
#endif

    // debugDiffCounter++;
    } else { // len2 == 0:
      ; 
    }

    // b. retain intersection if it is frequent
    if (maxdefectNew >= 0) {
#if DO_NOT_EXPAND_CLOSED_SETS
      if (maxdefectNew == maxdefect) {
	// cout << " found germ in by... " << endl;
	// closed set: do not expand, but remember for pattern output.
	rememberEquisupportExtension(bmIn->rowLabels[fromr]);
	assert(rowOut != NULL);
	delete[] rowOut; rowsOut[tor] = NULL;
      } else {
#endif
	rowLengthsOut[tor] = posOut;
	bmOut->rowLabels[tor] = bmIn->rowLabels[fromr];
	tor++;
	if (posOut > maxRowLength)
	  maxRowLength = posOut;
#if DO_NOT_EXPAND_CLOSED_SETS
      }
#endif
    } else {
      assert(rowOut != NULL);
      delete[] rowOut; rowsOut[tor] = NULL;
    }
  }

  bmOut->numRows = tor; // numRowsOut;
  bmOut->sizeRows = tor; // numRowsOut;
  bmOut->maxRowLength = maxRowLength;

#if DEBUG_LEVEL >= 20
  cout << "- done" << endl;
#endif
  return bmOut;
}




/**
 * Intersect all rows (with higher index) with a given row. 
 * @param bmIn original bit matrix.
 * @param row row to intersect all other rows (with higher index) with
 */
SparseBitmatrix* intersectSparseBitmatrix_dynamicMemory(SparseBitmatrix* bmIn, int row, int minsup) {
  assert(row >= 0 && row < bmIn->numRows);

  // cout << "intersect row1 = " << row << " (numRows = " << bmIn->numRows << ")" << endl;
  // 1. reset structure of bmOut.
  int numRowsIn = bmIn->numRows;
  int** rowsIn = bmIn->rows;
  int* rowLengthsIn = bmIn->rowLengths;
  int numRowsOut = numRowsIn - row - 1;
  SparseBitmatrix* bmOut = new SparseBitmatrix(numRowsOut);
  assert(numRowsOut <= bmOut->sizeRows);
  int** rowsOut = bmOut->rows;
  int* rowLengthsOut = bmOut->rowLengths;

  // 2. do the intersection:
  int* row1 = rowsIn[row];
  int len1 = rowLengthsIn[row];
  int tor = 0;
  int* rowOut = rowsOut[0];
  int maxRowLength = 0;
  for (int fromr = row+1; fromr < numRowsIn; fromr++) {
    assert(tor >= 0 && tor < numRowsOut);

    int* row2 = rowsIn[fromr];
    int len2 = rowLengthsIn[fromr];

    int maxLen = len2 < len1 ? len2 : len1;
    rowsOut[tor] = new int[maxLen]; bmOut->rowSizes[tor] = maxLen; 
    rowOut = rowsOut[tor];
    assert(rowOut != NULL);
    assert(rowsOut[tor] != NULL);
    assert(bmOut->rowSizes[tor] >= len2 || bmOut->rowSizes[tor] >= len1);

#if USE_ROW_SWAPPING    
    bool swapped = len1 < len2;
    if (swapped) {
      int* tmp = row1; row1 = row2; row2 = tmp;
      int tmplen = len1; len1 = len2; len2 = tmplen;
    }
#else
#define swapped false
#endif

    int pos1 = 0, pos2 = 0, posOut = 0;
    int item1 = row1[0];

#if USE_EARLY_EXIT    
    // int ln1 = len1, ln2 = len2;
    int ln1 = len1 - minsup + 1, ln2 = len2 - minsup + 1;
    // int ln1_0 = ln1, ln2_0 = ln2;
    // int ln1 = len1 - minsup, ln2 = len2 - minsup;
#endif

    assert(fromr < bmIn->sizeRows);
    assert(swapped ? row1 == bmIn->rows[fromr] : row2 == bmIn->rows[fromr]);
    assert(swapped ? pos1 < bmIn->rowSizes[fromr] : pos2 < bmIn->rowSizes[fromr]);
    int item2 = row2[0];

    // a. compute intersection:
    while (true) {
      if (item1 < item2) {
	pos1++; 
#if USE_EARLY_EXIT
	if (pos1 >= ln1) break;
#else
	if (pos1 >= len1) break;
#endif
	item1 = row1[pos1];
      } else if (item1 > item2) {
	pos2++; 
#if USE_EARLY_EXIT
	if (pos2 >= ln2) break;
#else
	if (pos2 >= len2) break;
#endif
	item2 = row2[pos2];
      } else { // if (item1 == item2) {
	assert(posOut >= 0);
	assert(posOut < len1); 
	assert(posOut < len2); 
	assert(tor >=0 && tor < bmOut->sizeRows);
	// if (posOut >= bmOut->rowSizes[tor]) cout << "FAIL: posOut = " << posOut << ", rowSizes=" << bmOut->rowSizes[tor] << endl;
	assert(posOut < bmOut->rowSizes[tor]);
	rowOut[posOut] = item1;
	posOut++;
#if USE_EARLY_EXIT
	pos1++;
	pos2++; 
	if (posOut < minsup) {
	  ln1++;
	  ln2++;
	} 
	if (pos1 >= ln1) break;
	if (pos2 >= ln2) break;
#else
	pos1++; if (pos1 >= len1) break;
	pos2++; if (pos2 >= len2) break;
#endif
	item1 = row1[pos1]; item2 = row2[pos2];
      }
    }

#if USE_ROW_SWAPPING    
    if (swapped) {
      row1 = row2; 
      len1 = len2;
    }
#endif

#if DEBUG_LEVEL >= 20
    cout << "intersect" << endl 
	 << "  ref.row " << row << ":";
    for (int k = 0; k < len1; k++)
      cout << " " << row1[k];
    cout << endl
	 << "  run.row " << fromr << ":";
    for (int k = 0; k < len2; k++)
      cout << " " << row2[k];
    cout << endl
	 << " " << (posOut >= minsup ? "+" : "-")
	 << "result  " << tor << ":";
    for (int k = 0; k < posOut; k++)
      cout << " " << rowOut[k];
    cout << endl;
#endif

    // b. retain intersection if it is frequent
    if (posOut >= minsup) {
#if DO_NOT_EXPAND_CLOSED_SETS
      if (posOut == len1) {
	// closed set: do not expand, but remember for pattern output.
	rememberEquisupportExtension(bmIn->rowLabels[fromr]);
	delete[] rowsOut[tor]; rowsOut[tor] = NULL;
      } else {
#endif
	rowLengthsOut[tor] = posOut;
	bmOut->rowLabels[tor] = bmIn->rowLabels[fromr];
	tor++;
	if (posOut > maxRowLength)
	  maxRowLength = posOut;
#if DO_NOT_EXPAND_CLOSED_SETS
      }
#endif
    } else {
      delete[] rowsOut[tor]; rowsOut[tor] = NULL;
    }
  }

  bmOut->numRows = tor; // numRowsOut;
  bmOut->sizeRows = tor; // numRowsOut;
  bmOut->maxRowLength = maxRowLength;

#if DEBUG_LEVEL >= 20
  cout << "- done" << endl;
#endif
  return bmOut;
}




