/*
 * MemoryAdhoc.h
 *
 * Time-stamp: <04/09/09 09:11:22 lars>
 *
 * history: 2004/04/01  1.0  LST created.
 */

#ifndef __MEMORY_ADHOC__
#define __MEMORY_ADHOC__

#include "SparseBitmatrix.h"

/**
 * Compute the difference of every other row (with higher index) minus a given row.
 * I.e., compute bmIn_row' \setminus bmIn_row for all row' > row.
 * @param bmIn original bit matrix.
 * @param row row to compute rest by.
 * @param minsup minimal support.
 */
SparseBitmatrix* setdifferenceBySparseBitmatrix_dynamicMemory(SparseBitmatrix* bmIn, int row, int maxdefect);

/**
 * Intersect all rows (with higher index) with a given row. 
 * @param bmIn original bit matrix.
 * @param row row to intersect all other rows (with higher index) with
 */
SparseBitmatrix* intersectSparseBitmatrix_dynamicMemory(SparseBitmatrix* bmIn, int row, int minsup);


// auto-expanding arrays:

#define DECLARE_ARRAY(array, capacity) \
  { \
    int array ## Capacity = capacity; \
    int array ## Length = 0; \
    int* array = new int[array ## Capacity]; \
  }

#define ARRAY_LENGTH(array) array ## Length
#define ARRAY_CAPACITY(array) array ## Capacity

#define APPEND_TO_ARRAY(array, value)	\
  { \
    if (array ## Length == array ## Capacity) { \
      array ## Capacity *= 2; \
      int* __old = array, array = new int[ array ## Capacity ]; \
      for (int __i = 0; __i < array ## Length; __i++) \
	array[__i] = __old[__i]; \
      delete[] __old; \
      array[array ## Length++] = value; \
  }


#endif
