/*
 * SparseBitmatrix.h
 *
 * Time-stamp: <04/09/04 13:32:50 lars>
 *
 * history: 2004/04/01  1.0  LST created.
 */

#ifndef __SPARSE_BITMATRIX__
#define __SPARSE_BITMATRIX__ t

#include <iostream>
 
using namespace std;

/**
 * A sparse bit matrix, implemented as list of columns having 1 for each row. 
 *
 * @author Lars Schmidt-Thieme
 * @version 1.0
 */

class SparseBitmatrix {
 public:

  int numRows;
  // int numItems;
  int** rows;
  int maxRowLength;
  int* rowLengths;
  int* rowLabels;

  // size (i.e., allocated memory):
  int sizeRows;
  int* rowSizes;

  SparseBitmatrix(int numRows);
  SparseBitmatrix(int numRows, int rowLength);
  SparseBitmatrix(int numRows, int* rowLengths);
  SparseBitmatrix(int numRows, int* rowLengths, int* rowLabels);
  ~SparseBitmatrix();
  
  void dump();
  void dumpStructure();
  void sortRows();

  /**
   * Retain only rows with a given minimal row length.
   * (Re)computes maxRowLength as a side-effect.
   * @return number of rows retained.
   */
  int SparseBitmatrix::filterMinimalRowLength(int minRowLength);

  int SparseBitmatrix::filterMinimalRowLength_destroyStructure(int minRowLength);

  /**
   * Shift row storage s.t. given row is row 0.
   * Release memory for all rows before. 
   */
  void SparseBitmatrix::shiftRowStorage(int firstRow);	

#if USE_DYNAMIC_MEMORY_NODE_EXIT
 private:
  int rowshift;
#endif
};

#endif
